/* 
 * File:   evaluator5.cpp
 * Article: http://www.saboti.nu/writing-an-expression-evaluator/index.html
 *
 * AGPL3
 */
 
/*
 * Left to right parser
 * Parentheses
 * Precedence
 * Power operator
 * Associativity
 */

#include <cstdlib>
#include <iostream>
#include <cmath>

enum class Operator
{
	Null,
	Addition,
	Subtraction,
	Multiplication,
	Division,
	Power,
};

enum class Associativity
{
	Left,
	Right,
};

bool read_expression(char const *& pos, int & result, Operator previous_op = Operator::Null);

bool is_digit(char const c)
{
	return c >= '0' && c <= '9';
}

bool read_number(char const *& pos, int & num)
{
	if(*pos == '(')
	{
		++pos;
		if(!read_expression(pos, num) || *pos != ')')
		{
			return false;
		}
		++pos;
		return true;
	}
	
	num = 0;
	if(is_digit(*pos))
	{
		do
		{
			num *= 10;
			num += *pos - '0';
			++pos;
		} while(is_digit(*pos));
		return true;
	}
	return false;
}

bool read_operator(char const *& pos, Operator & op)
{
	switch(*pos)
	{
		case '+':
			op = Operator::Addition;
			break;
		case '-':
			op = Operator::Subtraction;
			break;
		case '*':
			op = Operator::Multiplication;
			break;
		case '/':
			op = Operator::Division;
			break;
		case '^':
			op = Operator::Power;
			break;
		default:
			return false;
	}
	++pos;
	return true;
}

int p(Operator op)
{
	switch(op)
	{
		case Operator::Addition:
		case Operator::Subtraction:
			return 10;
		case Operator::Multiplication:
		case Operator::Division:
			return 20;
		case Operator::Power:
			return 30;
		default:
			return 0;
	}
}

Associativity a(Operator op)
{
	switch(op)
	{
		case Operator::Power:
			return Associativity::Right;
		default:
			return Associativity::Left;
	}
}

bool compute_binary(Operator op, int lhs, int rhs, int & result)
{
	switch(op)
	{
		case Operator::Addition:
			result = lhs + rhs;
			break;
		case Operator::Subtraction:
			result = lhs - rhs;
			break;
		case Operator::Multiplication:
			result = lhs * rhs;
			break;
		case Operator::Division:
			result = lhs / rhs;
			break;
		case Operator::Power:
			result = std::pow(lhs, rhs);
			break;
		default:
			return false;
	}
	return true;
}

bool read_expression(char const *& pos, int & result, Operator previous_op)
{
	if(read_number(pos, result))
	{
		Operator op;
		while(read_operator(pos, op))
		{
			/*
			 * Precedence check.
			 */
			bool op_less_previous = p(op) < p(previous_op);
			if(op_less_previous)
			{
				--pos;
				break;
			}
			/*
			 * Associativity check.
			 */
			bool op_equal_previous = p(op) == p(previous_op);
			bool op_left_associativity = a(op) == Associativity::Left;
			if(op_equal_previous && op_left_associativity)
			{
				--pos;
				break;
			}
			int num;
			if(!read_expression(pos, num, op))
			{
				return false;
			}
			if(!compute_binary(op, result, num, result))
			{
				return false;
			}
		}
		return true;
	}
	return false;
}

bool calculate(char const *& pos, int & result)
{
	if(!pos)
	{
		return false;
	}
	return read_expression(pos, result) && !*pos;
}

int main(int argc, char const ** argv)
{
	
	if(!argv[1])
	{
		std::cout << "You must invoke this program with an expression\n";
		return 1;
	}
	
	char const * text = argv[1];
	int result;
	
	if(calculate(text, result))
	{
		std::cout << result << std::endl;
	}
	else
	{
		std::cout << "There seems to be an error here: '" << text << "'\n";
		return 1; 
	}
	
	return 0;
}

